from re import findall, match
from tqdm import tqdm
from time import sleep
import requests
from random import randint, choice
from aiohttp import ClientSession
from datetime import datetime
from json import loads, dumps, decoder
from asyncio import gather, get_event_loop
from stringcolor import *

from sterra._userAgents_ import USER_AGENTS
from sterra._sterrage_ import ENDPOINTS_TEST_LIST
from sterra._outputs_ import LOGOS, _no_colors
from sterra.SterraException import LoginError, RateLimitError, UserNotFoundError, PrivateAccError, NoFollowError, MutualsError

USER_AGENT = choice(USER_AGENTS["safari"])
IOS_UAGENT = choice(USER_AGENTS["ios"])
global _
_ = None

def _credToSessID(creds:list) -> tuple[str or bool, str or bool]:
    """Returns the sessionid of the credential given"""
    u, p = tuple(creds)

    session = requests.Session()
    session.headers.update({
        "Accept-Encoding": "gzip, deflate",
        "Accept-Language": "en-US,en;q=0.8",
        "Connection": "keep-alive",
        "Content-Length": "0",
        "Host": "www.instagram.com",
        "Origin": "https://www.instagram.com",
        "Referer": "https://www.instagram.com/",
        "User-Agent": USER_AGENT,
        "X-Instagram-AJAX": "7a3a3e64fa87",
        "X-Requested-With": "XMLHttpRequest"
        })

    mainPage = session.get("https://www.instagram.com")
    csrf = findall(r"csrf_token\":\"(.*?)\"", mainPage.text)[0]
    if not csrf:
        return False, "No csrf token found. Wait 1 hour and retry, or change of account."
    
    session.headers.update({"x-csrftoken": csrf})
    sleep(randint(1, 2))

    data = {"username": u, "enc_password": "#PWD_INSTAGRAM_BROWSER:0:{}:{}".format(int(datetime.now().timestamp()), p)}

    postResp = session.post("https://www.instagram.com/accounts/login/ajax/", data=data, allow_redirects=True)
    response_data = postResp.json()

    if "two_factor_required" in response_data:
        _.r(LoginError("Disable 2-factor authentication to login."))
    if "message" in response_data and response_data["message"] == "checkpoint_required":
        _.r(LoginError("Check Instagram app for a security confirmation."))
    try:
        if not response_data["authenticated"]:
            _.r(LoginError("Invalid credentials."))
    except KeyError:
        if response_data["spam"]:
            _.r(RateLimitError())
    else:
        _.p(dumps(response_data,indent=4))
        _.r(LoginError("Unknown error. Please report the upper dictionnary to the issue section of Sterra."))

    return postResp.cookies["sessionid"]

def _idToUsername(i:str,ss:str) -> str or bool:
    """Returns the username associated to the instagram id
    -i: id, -s: sessonid"""
    apiResp = requests.get(f'https://i.instagram.com/api/v1/users/{i}/info/', headers={'User-Agent':IOS_UAGENT}, cookies = {'sessionid':ss}, allow_redirects=False)
    try:
        apiJson = loads(apiResp.text)
        return apiJson['user']['username']
    except decoder.JSONDecodeError:
        return None if apiResp.status_code == 404 else False

def _targListToHashesAndViceV(s:str) -> list:
    """Returns the hashes for the username list to scrape
    - t: target list"""
    repère = {
        "followers":["c76146de99bb02f6415203be841dd25a"],
        "following":["d04b0a864b4b54837c0d870b0e77e076"],
        "both":["c76146de99bb02f6415203be841dd25a","d04b0a864b4b54837c0d870b0e77e076"],
        "mutuals":["c76146de99bb02f6415203be841dd25a","d04b0a864b4b54837c0d870b0e77e076"]
    }
    for k, v in repère.items():
        if s in [k]+v:
            return (k if s in v else v)

class _instagram:
    def __init__(self, o:object, orgtar:str ,**kwargs:dict) -> None:
        global _
        _ = o
        self.original_target = orgtar

        self.default_h = {'User-Agent':USER_AGENT}
        sessionid = kwargs.get("login_session_id")
        if not sessionid:
            sessionid = _credToSessID(kwargs.get("login_credentials"))

        else:
            self._checkSessidFormat(sessionid)
            self._checkvalidSessid({"sessionid":sessionid})
        self.cookies = {"sessionid":sessionid}
        self.used_account = sessionid.split("%")[0]

        username = kwargs.get("username")
        if not username:
            username = _idToUsername(kwargs.get("id"),sessionid)
            if not username:
                _.r((UserNotFoundError("username" if kwargs.get("username") else "id") if username is None else RateLimitError))
        self.username = username
        
        __a1 = requests.get(f'https://www.instagram.com/{username}/channel/?__a=1', cookies=self.cookies, headers={'User-Agent':USER_AGENT}, allow_redirects=False)
        st = __a1.status_code
        if st != 200:
            _.r((UserNotFoundError("username" if kwargs.get("username") else "id") if st == 404 else RateLimitError))
        j = __a1.json()
        if not j:
            _.r(UserNotFoundError("username" if kwargs.get("username") else "id"))
        j = j['graphql']['user']
        self.acc_infos = {'id': j['id'], 'followers': j['edge_followed_by']['count'], 'following': j['edge_follow']['count']}
        self.target_ig_id = self.acc_infos["id"]

        if j['is_private'] and int(j['edge_owner_to_timeline_media']['count']):
            self._verifyPrivAccAccess()

        _.p(f"""Target > {str(bold(self.username))} ({self.acc_infos["followers"]} followers, {self.acc_infos["following"]} following)""",logo="x")

        self.target_list:str = kwargs.get("target") # ["followers", "following", "both", "mutuals"]

        # relger ce truc tu possible express désactivé meme si la seul liste target est courte
        self.delay = kwargs.get("delay") if kwargs.get("delay") else 0.0
        self.all_infos = kwargs.get("all_infos")
        
        self.no_exp_limit = kwargs.get("no_exp_limit")
        self.want_express = kwargs.get("express")
        self.express = self._canExpress()
        
    def __call__(self) -> dict:
        return vars(self)

    def _canExpress(self, lenght:int=None) -> bool or None:
        if self.want_express and self.no_exp_limit:
            return True

        if self.original_target == "followers":
            possible_express = True if self.acc_infos['followers'] < 109 else False
        elif self.original_target == "following":
            possible_express = True if self.acc_infos['following'] < 109 else False
        elif self.original_target == "mutuals":
            if lenght:
                possible_express = True if lenght < 109 else False
            else:
                possible_express = None
        else:
            possible_express = True if (self.acc_infos['followers']+self.acc_infos['following']) < 109 else False

        if possible_express is None:
            return None
        return (True if self.want_express and possible_express else False)

    def _checkSessidFormat(self,s) -> None:
        """Compares the sessionId to a regex and return True if it matches"""
        if not match(r'[0-9]{1,32}%[0-9a-zA-Z]{16}%[0-9A-Z]{3,4}',s):
            _.r(LoginError("Invalid sessid format"))

    def _checkvalidSessid(self,c:dict) -> None:
        """Raises loginError if the sessionId is not working"""
        status = requests.get(f"https://www.instagram.com/{choice(ENDPOINTS_TEST_LIST)}",headers=self.default_h,cookies=c,allow_redirects=False).status_code
        if status != 200:
            _.r(LoginError("Invalid sessid"))

    def _verifyPrivAccAccess(self) -> None:
        """Works only if account has posts"""
        page = requests.get(f"https://www.instagram.com/{self.username}/",headers=self.default_h,cookies=self.cookies,allow_redirects=False).text
        try:
            infos = loads(page.split("<script type=\"text/javascript\">window._sharedData = ")[-1].split(";</script>")[0])["entry_data"]["ProfilePage"][0]["graphql"]["user"]
            if not infos["edge_owner_to_timeline_media"]["edges"]+infos["edge_felix_video_timeline"]["edges"]:
                _.r(PrivateAccError(self.username))
        except KeyError:
            _.p("There might be an issue with json response in stegram:_verifyPrivAccAccess, please report it to the issue section. This issue isn't fatal.")
        else:
            pass
    
    def _followListScraper(self,h:str,has_next_page=True,act_attempts=0,att=3) -> list[str]:
        """Scrapes the follow list (precised with the hash) of the target username
        Only "h" matters: hash"""
        targlist = _targListToHashesAndViceV(h)
        total_follow = self.acc_infos[targlist]

        pbar = tqdm(total=total_follow, desc=f"""{LOGOS["Python"] if _.colors else _no_colors(LOGOS["Python"])} {targlist.capitalize()} extraction""")
        follow_list = []

        var = {"id":self.target_ig_id,"first":50}
        params = {"query_hash":h,"variables":dumps(var)}

        def resolver(json_resp:dict) -> bool:
            json_resp = json_resp["data"]["user"]["edge_followed_by" if h == "c76146de99bb02f6415203be841dd25a" else "edge_follow"]

            for node in json_resp["edges"]:
                pbar.update(1)
                follow_list.append(node["node"]["username"])

            if json_resp["page_info"]["has_next_page"]:
                var["after"] = json_resp["page_info"]["end_cursor"]
                return True
            return False

        while has_next_page and act_attempts < att:
            queryAPIResp = requests.get("https://www.instagram.com/graphql/query/", params=params, cookies=self.cookies)
            if not queryAPIResp.cookies.get_dict():
                _.r(PrivateAccError(self.username))
            elif queryAPIResp.status_code == 200:
                act_attempts = 0
                has_next_page = resolver(queryAPIResp.json())
                if has_next_page:
                    params["variables"] = dumps(var)

            else:
                act_attempts += 1
                _.p(f"{act_attempts}/{att} attempts, sleeping 10 sec")
                sleep(10.0)

        pbar.close()
        if has_next_page: # Travail est pas fini, donc qu'il y a eu une couille
            _.r(RateLimitError("Rate limit happenning during list of follows scraping. Wait a bit, or change of login account."))

        return follow_list

    def scrapeTargetLists(self) -> tuple[list,list]:
        """Returns the list to scrape depending on both or not, and on the data already known to block every error that could happen in the process"""
        follow_counts_list = [self.acc_infos[k] for k in ['followers','following']]
        if 0 in follow_counts_list:
            if follow_counts_list == [0,0]:
                _.r(NoFollowError())
            empty_usernames_list_name = [k for k in ['followers','following'] if self.acc_infos[k] == 0][0]

            if self.target_list == "mutuals":
                _.r(MutualsError("No mutuals possible since followers or following list is empty."))
            elif self.target_list == "both":
                _.r(NoFollowError(f"""\"Both" list can't be retrieved since your target has no {empty_usernames_list_name}."""))
        
            elif self.target_list == empty_usernames_list_name:
                _.r(NoFollowError(f"{self.target_list.capitalize()} can't be scraped since it is empty"))
            
        scraping_hashes = _targListToHashesAndViceV(self.target_list)
        follow_usernames = []
        for hash in scraping_hashes:
            follow_usernames.append(self._followListScraper(hash))

        return tuple(follow_usernames)

    def distractAPI(self) -> None: # Not used for now
        """Function made to call a random instagram API other than /USERNAME/?__a/1 to "simulate" very badly an human doing something else.
        These endpoints are sensitive;
        - some of the responses contains personnal informations such as login places, phone numbers, email adresses, ...
        - some of the endpoints can also change your account parameters.
        
        NO DATA IS SAVED, WE ONLY GET THE STATUS CODE OF THE RESPONSE
        NO DATA IS POSTED, ALL REQUESTS ARE get REQUESTS"""
        endpoints = ENDPOINTS_TEST_LIST
        requests.get(f"""https://www.instagram.com/{choice(endpoints)}""", headers=self.default_h, cookies=self.cookies).close()
    
    def getUsernameDetails(self,l:list,already:list=None) -> list:
        """Returns list of ?__a=1 details on each accounts
        -l:list of account, -already:part already exported"""
        self.express = self._canExpress(len(l)) if self.express is None else self.express
        # Vu que la liste de mutuels est calculée après le __init__, on actualise le self.express en fonction de la longueur de la liste.

        done = already if already else []
        lWithoutDone = [x for x in l if x not in [y["username"] for y in already]] if already else l

        if not self.express and len(l) > 20:
            _.p(f"""Exportation in normal mode can take a while. It is normal. {str(underline("You are free to do something else during this time"))}.""", logo="i")

        pbar = tqdm(total=len(l), desc=f"""{LOGOS["Instagram"] if _.colors else _no_colors(LOGOS["Instagram"])} Getting users details""")
        pbar.update(len(done))

        def outputDict(d:dict,username:str=None) -> dict:
            """Converts the raw dict to a selected informations dict"""
            ret = {}
            isNotFound = False
            try:
                d = d['graphql']['user']
                d["link"] = f'''https://www.instagram.com/{d["username"]}/'''
            except KeyError:
                isNotFound = True


            if isNotFound:
                classicKeys = ["id","username","full_name","link","biography","is_private","followers","following","posts","external_url","is_business_account","is_professional_account","is_verified"]
                allInfoKeys = ['business_address_json','business_category_name','business_contact_method','business_email','business_phone_number','connected_fb_page','edge_mutual_followed_by','count','has_ar_effects','has_channel','has_clips','has_guides','hide_like_and_view_counts','is_joined_recently']
                for elemPath in (classicKeys+allInfoKeys if self.all_infos else classicKeys):
                    ret[elemPath] = username if elemPath == "username" else None

            else:
                # Paths à aller chercher dans le Json brut
                classicKeys = [['id'],['username'],['full_name'],['link'],['biography'],['is_private'],['edge_followed_by','count'],['edge_follow','count'],['edge_owner_to_timeline_media','count'],['external_url'],['is_business_account'],['is_professional_account'],['is_verified']]
                allInfoKeys = [['business_address_json'],['business_category_name'],['business_contact_method'],['business_email'],['business_phone_number'],['connected_fb_page'],['edge_mutual_followed_by','count'],['has_ar_effects'],['has_channel'],['has_clips'],['has_guides'],['hide_like_and_view_counts'],['is_joined_recently']]

                # Changement du nom de la nouvelle clée si ancienne première clée du path présent dans ce dict:
                custom_Keys = {'edge_followed_by':'followers','edge_follow':'following','edge_owner_to_timeline_media':'posts','edge_mutual_followed_by':'mutual_followed_by_count'}

                for elemPath in (classicKeys+allInfoKeys if self.all_infos else classicKeys): # Pour élément dans liste de path par défaut (+all_infos si self.all_infos)
                    c = custom_Keys[elemPath[0]] if elemPath[0] in list(custom_Keys.keys()) else False # Changement de clé = False si pas dans le dict de remplacement de clées
                    j = d
                    for key in elemPath:
                        j = j[key]
                        if key == elemPath[-1]: # Lorsque l'on arrive au bout du path, on enregistre dans le dict à sortir
                            ret[c if c else key] = j
            
            return ret

        def normalExtract(ret:list) -> list:
            """Calls to ?__a=1 for each username in normal speed"""
            with requests.Session() as s:
                for i, u in enumerate(lWithoutDone):
                    rep = s.get(f'https://www.instagram.com/{u}/channel/?__a=1',headers={'User-Agent':USER_AGENT,'Referer':f'https://www.instagram.com/{u}'},cookies=self.cookies,allow_redirects=False)
                    sleep(self.delay)
                    if rep.status_code == 200:
                        ret.append(outputDict(loads(rep.text)))
                        pbar.update(1)

                    elif rep.status_code == 404:
                        ret.append(outputDict({},u))
                        
                    # if i > 3: break # volontary rate limit for tests

            return ret

        async def expressExtract(ret:list) -> list:
            """Calls to ?__a=1 for each username in asynchronous speed"""
            async with ClientSession() as s:
                async def fetch(u:str):
                    """Process the request and append the result"""
                    async with s.get(f'https://www.instagram.com/{u}/channel/?__a=1',headers={'User-Agent':USER_AGENT,'Referer': f'https://www.instagram.com/'},cookies=self.cookies,allow_redirects=False) as __a1:
                        if __a1.status == 200:
                            ret.append(outputDict(await __a1.json()))
                            pbar.update(1)
                        
                        elif __a1.status == 404:
                            ret.append(outputDict({},u))

                rq_corro = []
                for u in lWithoutDone:
                    rq_corro.append(fetch(u))
                futures = gather(*rq_corro)
                await futures
                await s.close()
            
            return ret

        rtr = get_event_loop().run_until_complete(expressExtract(done)) if self.express else normalExtract(done)
        pbar.close()
        return rtr


from re import findall

# Coeff:
_PERSONNAL_PRIVATE_IN_DESC = 2
_PERSONNAL_PRIVATE_IN_NAME = 3
_PERSONNAL_IS_PRIVATE = 6
_PERSONNAL_FOLLOWERS_COUNT = 4
_PERSONNAL_FOLLOWING_COUNT = 1
_PERSONNAL_EXTERNAL_URL = 1
_PERSONNAL_IS_PROFESSIONAL = 3
_PERSONNAL_IS_BUSINESS = 3
_PERSONNAL_IS_VERIFIED = 3
_PERSONNAL_POSTS = 3

_INTERESTS_PRIVATE_IN_DESC = 1
_INTERESTS_PRIVATE_IN_NAME = 1
_INTERESTS_IS_PRIVATE = 7
_INTERESTS_FOLLOWERS_COUNT = 7
_INTERESTS_FOLLOWING_COUNT = 1
_INTERESTS_EXTERNAL_URL = 3
_INTERESTS_IS_PROFESSIONAL = 4
_INTERESTS_IS_BUSINESS = 5
_INTERESTS_IS_VERIFIED = 3
_INTERESTS_POSTS = 2

class _analysis:
    def __init__(self,_:object, data:list, **kwargs) -> None:
        self._ = _
        self.data = data
        
        self.invert = kwargs.get("invert")
        self.ignore_over = kwargs.get("ignore_over") if kwargs.get("ignore_over") else 10000
        self.pctg = kwargs.get("pctg") if kwargs.get("pctg") else 0
        self.size = kwargs.get("size") if kwargs.get("size") else len(self.data)

    def __call__(self, target:str) -> list:
        return self._order(getattr(self, target)())

    def _order(self, l:list):
        order = sorted(l, key=lambda item: item['%'], reverse=self.invert)
        order = order[-self.size:]
        return [x for x in order if x['%'] >= self.pctg]

    def _dictFormat(self, d:dict) -> dict:
        rtr = {}
        for k, v in d.items():
            if v in ["True","False","None"]:
                v = True if v == "True" else False if v == "False" else None
            if k in ["posts","followers","following","mutual_followed_by_count"]:
                rtr[k] = 0 if v is None else int(v)
            else:
                rtr[k] = v
        return rtr

    def personnal(self) -> list:
        """Data analysis to get probas of personnal accounts"""
        def fais_le_boulot(acc:dict) -> dict:
            """Objective probabilities"""
            probas = []
            repv = r'([pP]{1,4}[rR]{1,4}([iI]{0,4}[vV]{1,4}[aAéÉeE]{1,10}[tT]{0,4}[eE]{0,4}|[vV]{1,4}[tT]{1,4}[eE]{0,9})|[pP][vV]{1,4})' # PRIVATE / PRIVÉ
            
            if acc['biography']:
                if findall(repv, acc['biography']):
                    probas += [100]*_PERSONNAL_PRIVATE_IN_DESC
            
            if acc["full_name"]:
                if findall(repv, acc["full_name"]):
                    probas += [100]*_PERSONNAL_PRIVATE_IN_NAME
            
            probas += [100 if acc['is_private'] else 0]*_PERSONNAL_IS_PRIVATE
            probas += [100 if acc['followers'] < (self.ignore_over / 15) else (50 if acc['followers'] < self.ignore_over else 0)]*_PERSONNAL_FOLLOWERS_COUNT
            probas += [100 if acc['following'] < 50 else (50 if acc["following"] < 500 else 0)]*_PERSONNAL_FOLLOWING_COUNT
            probas += [0 if acc['external_url'] else 100]*_PERSONNAL_EXTERNAL_URL

            """Subjective probbailities"""
            sub = 0
            for i in probas:
                sub += i
            SUB = int(sub / len(probas))
            
            probas += [100 if not acc['is_professional_account'] else (50 if SUB > 75 else 0)]*_PERSONNAL_IS_PROFESSIONAL
            probas += [100 if not acc['is_business_account'] else (50 if SUB > 75 else 0)]*_PERSONNAL_IS_BUSINESS
            probas += [100 if not acc['is_verified'] else (100 if SUB > 90 else 0)]*_PERSONNAL_IS_VERIFIED
            probas += [100 if acc['posts'] > 50 and SUB > 85 else 0]*_PERSONNAL_POSTS

            final = 0
            for p in probas:
                final += p
            final /= len(probas)
            final = final if final < 100 else 99

            retour = {"%": final}
            retour.update(acc)
            return retour

        return [fais_le_boulot(self._dictFormat(acc)) for acc in self.data if acc["id"]]

    def interests(self) -> list:
        """"""
        def fais_le_boulot(acc:dict) -> dict:
            """Objective probabilities"""
            probas = []
            repv = r'([pP]{1,4}[rR]{1,4}([iI]{0,4}[vV]{1,4}[aAéÉeE]{1,10}[tT]{0,4}[eE]{0,4}|[vV]{1,4}[tT]{1,4}[eE]{0,9})|[pP][vV]{1,4})' # PRIVATE / PRIVÉ
            
            if acc['biography']:
                if findall(repv, acc['biography']):
                    probas += [0]*_INTERESTS_PRIVATE_IN_DESC
            
            if acc["full_name"]:
                if findall(repv, acc["full_name"]):
                    probas += [0]*_INTERESTS_PRIVATE_IN_NAME

            probas += [0 if acc['is_private'] else 100]*_INTERESTS_IS_PRIVATE
            probas += [100 if acc['followers'] > self.ignore_over else (50 if acc['followers'] > (self.ignore_over/2) else 0)]*_INTERESTS_FOLLOWERS_COUNT
            probas += [100 - ((acc['following'] if acc['following'] < 1500 else 1500) * 100) / 1500]*_INTERESTS_FOLLOWING_COUNT
            probas += [0 if acc['external_url'] else 100]*_INTERESTS_EXTERNAL_URL
            probas += [100 if not acc['is_verified'] else 0]*_INTERESTS_IS_VERIFIED
            probas += [100 if acc['posts'] > 50 else 0]*_INTERESTS_POSTS

            """Subjective probbailities"""
            sub = 0
            for i in probas:
                sub += i
            SUB = int(sub / len(probas))
            
            probas += [100 if acc['is_professional_account'] else (50 if SUB > 75 else 0)]*_INTERESTS_IS_PROFESSIONAL
            probas += [100 if acc['is_business_account'] else (50 if SUB > 75 else 0)]*_INTERESTS_IS_BUSINESS

            final = 0
            for p in probas:
                final += p
            final /= len(probas)
            final = final if final <= 100 else 99

            retour = {"%": final}
            retour.update(acc)
            return retour

        return [fais_le_boulot(self._dictFormat(acc)) for acc in self.data if acc["id"]]


from sterra._sterrage_ import HISTORY_PATH,CLEARED_HISTORY,PARTS_PATH
from json import loads, dumps
from time import time
from re import findall
from os import remove
from stringcolor import *

ACTUAL_HISTORY:dict = loads(open(HISTORY_PATH,"r").read())
PDICT = ACTUAL_HISTORY["part"]

def _mkId() -> str:
    return str(time()).replace(".","")

def all(**kwargs) -> dict:
    """Returns all the history under dict"""
    return ACTUAL_HISTORY

def clear(**kwargs) -> str:
    """Returns True when the history has been cleared"""
    if input("You are about to clear your export history. Y to confirm, other to abort.\n -> ").lower() == "y":
        with open(HISTORY_PATH,"w") as w:
            w.write(CLEARED_HISTORY)
        return "Export history cleared."
    return "Export history cleaning failed."

def file_id(fileid:str=None,**kwargs) -> str:
    """Returns the path associated to an id"""
    Id = fileid if fileid else kwargs.get("file_id")
    if Id:
        for k, v in ACTUAL_HISTORY.items():
            if k == Id:
                return f"""Path associated to id {Id}:\n    {v["path"]}""" if kwargs.get("tostr") else v["path"]
    return f"Path associated to id {Id} found." if kwargs.get("tostr") else None

def match(**kwargs) -> dict:
    """Returns elements of history matching with input"""
    Match = kwargs.get("match")
    isRegex = Match[1:-1] if "??" == Match[0]+Match[-1] else False
    matching = {}
    for k, v in ACTUAL_HISTORY.items():
        if k == "part":
            pass
        else:
            if isRegex and findall(f"(?:{isRegex})", v["path"]):
                matching[k] = v["path"]
            elif Match in v["path"]:
                matching[k] = v["path"]
    return (f"""Paths matching with {f'regex r"{isRegex}"' if isRegex else Match}:\n    """+"\n    ".join(matching) if matching else ("Nothing found.", "Exception")) if kwargs.get("tostr") else matching

def path(**kwargs) -> str:
    """Returns the id associated to a path"""
    Path = kwargs.get("path")
    for k, v in ACTUAL_HISTORY.items():
        if k != "part":
            if v["path"] == Path:
                return f"Id for path {Path}:\n    {k}" if kwargs.get("tostr") else k
    return f"No id associated to the path {Path} found.", "Exception" if kwargs.get("tostr") else None

def add(**kwargs) -> str:
    """Write down the input dict and returns the id of the stored data"""
    Id = _mkId()
    ACTUAL_HISTORY[Id] = kwargs
    with open(HISTORY_PATH,"w") as w:
        w.write(dumps(ACTUAL_HISTORY,indent=4))
    return Id

def add_part(part_path:str=None, target:str=None) -> str:
    """Return the id of the part created or existing"""
    if part_path:
        return findall(r"[0-9]{16}",part_path)[-1]
    else:
        Id = _mkId()
        Path = f"{PARTS_PATH}{Id}.json"
        PDICT[Id] = {"path":Path,"target":target}
        with open(HISTORY_PATH,"w") as w:
            w.write(dumps(ACTUAL_HISTORY,indent=4))
        return Id, Path

def get_part(part_option:str=True) -> tuple[str,str]:
    if part_option is True:
        pKeys = list(PDICT.keys())
        return (pKeys[-1], PDICT[pKeys[-1]]["path"], PDICT[pKeys[-1]]["target"])
    try:
        return (part_option, PDICT[part_option]["path"], PDICT[part_option]["target"])
    except KeyError:
        if type(part_option) is str:
            Id = part_option.split("/")[-1].split(".")[0]
            return (Id, PDICT[Id]["path"], PDICT[Id]["target"])
    return None

def clear_parts(**kwargs) -> bool:
    if input("You are about to delete all your parts. Y to confirm, other to abort.\n -> ").lower() == "y":
        try:
            try:
                remove(PARTS_PATH)
            except PermissionError:
                return "Parts deletion failed, os.remove() raised a PermissionError.", "Exception"
            return "Parts deletion done."
        except FileNotFoundError:
            pass
    return "Parts deletion failed.", "Exception"

def compare_tree(**kwargs):
    def climb_the_tree(fileid) -> dict:
        file_path = file_id(**{"file_id":fileid})
        decomposed = exman()._decompose_path(file_path)
        if list(decomposed.keys()) != ["file_path","format","target"]:
            return {"base":decomposed["file_path"]}
        else:
            dcomp_split = decomposed["target"].replace("(","").replace(")","").split("&&" if "&&" in decomposed["target"] else "||")
            return {file_id(**{"file_id":item}):climb_the_tree(item) for item in dcomp_split}
                
    from sterra.exterra import exman
    return f"""History tree for compared file {kwargs["compare_tree"]}:\n"""+dumps({file_id(**{"file_id":kwargs["compare_tree"]}):climb_the_tree(kwargs["compare_tree"])}, indent=4)




from sterra.histerra import file_id, add_part, add, get_part, match
from sterra._sterrage_ import DEFAULT_EXPORT_PATH, DAY_STRF, HOUR_STRF
from datetime import date, datetime
from os.path import join, exists
from re import findall

REFDIC = dict(
    us="username",
    ta="target",
    da="date",
    ho="hour",
    li="list_id"
)

class exman:
    """exman is manage exports, rule names, type of export, etc ...
    The export was causing too much exceptions about names and shit, it wasn't possible to continue like this"""
    def __init__(self, _:object=None) -> None:
        self._ = _

    def _decompose_path(self, file_path:str) -> dict:
        """Decompose the path to get all the informations on the file"""
        retour = {
            "file_path": file_path
        }
        filenm = file_path.split("/")[-1] if "/" in file_path else file_path
        fsplit = filenm.split(".")
        retour["name"] = fsplit[0]
        retour["format"] = fsplit[-1]
        try:
            if "&&" in fsplit[0]:
                fsplit[0].replace("&&","@@")
            for k, v in [tuple(item.split(";")) for item in fsplit[0].split("&")]:
                retour[REFDIC[k]] = v.replace("@@","&&")
        
        except ValueError: # For names like "ta;(16441852344045281&&1644185302525507).csv" that doesn't have "&"
            if findall(r"[0-9]{8,20}\.", filenm) and not ";" in fsplit[0]: # For parts
                pass
            else:
                try:
                    k, v = fsplit[0].split(";")
                    retour[REFDIC[k]] = v
                except ValueError: # Custom names (e.g.: bleu.xlsx)
                    pass

        return retour

    def _make_pathfiles(self, path:str, name:str) -> str:
        """Makes the dirs if the pathfile doesn't exist"""
        def buildPath(path, name):
            newpath = join(path, name)
            try:
                open(newpath,"r").close()
                raise FileExistsError
            except FileNotFoundError:
                return newpath

        ext = name.split(".")[-1]
        newpath = None
        while not newpath:
            try:
                newpath = buildPath(path=path, name=name)
            except FileExistsError:
                self._.p(f"File name \"{name}\" already exists. Choose a new one:", logo="Interrogation")
                name = f"""{input("    -> ")}.{ext}"""

        if not exists(newpath):
            from os import makedirs
            try:
                makedirs("/".join(newpath.split("/")[:-1])+"/")
            except FileExistsError:
                pass
        Id = add(**dict(path=newpath))
        return Id, newpath

    def part(self, target_list:str, part_path:str=None) -> str:
        if part_path:
            exists = get_part(part_path)
            if type(exists) is tuple:
                return exists
            else:
                return add_part(target=target_list)
        else:
            return add_part(target=target_list)

    def custom_name(self, name:str, Format:str="xlsx", path:str=DEFAULT_EXPORT_PATH) -> str:
        """"""
        return self._make_pathfiles(path, f"{name}.{Format}")

    def classic_export(self, target:str, username:str, Format:str="xlsx", path:str=DEFAULT_EXPORT_PATH) -> str:
        """"""
        return self._make_pathfiles(path, f"""us;{username}&ta;{target}&da;{date.today().strftime(DAY_STRF)}&ho;{datetime.now().strftime(HOUR_STRF)}.{Format}""")

    def compare_export(self, file1:str, file2:str, target:str, Format:str="xlsx", path:str=DEFAULT_EXPORT_PATH) -> str:
        """"""
        file_ids = [list(match(**{"match":f}).keys())[0] for f in [file1, file2]]
        if None in file_ids:
            raise Exception("Couille ici")
        return self._make_pathfiles(path, f"""ta;({("&&" if target == "common_usernames" else "||").join(file_ids)}).{Format}""")
        
    def analysis_export(self, list_id:str, target:str, Format:str="xlsx", path:str=DEFAULT_EXPORT_PATH) -> str:
        """"""
        return self._make_pathfiles(path, f"""li;{list_id}&ta;{target}.{Format}""")
        # li == list_id

#+ ajouter ddate et heure a tous les noms ???


from stringcolor import *
from sterra._sterrage_ import VERSION
from random import choice

HORIZONTAL = "LightSlateBlue" #8787ff
VERTICAL = "Purple" #800080
OTHER = "DarkCyan" #00af87

def _generate_logo() -> str:
    # https://patorjk.com/software/taag/#p=display&h=2&f=Small%20Slant&t=sterra
    GIT_COM = "https://github.com/"
    NAME = "novitae"
    NOVI_URL = f"{GIT_COM}{NAME}"
    VERSION_POSITON = 20

    _LOGO = """           __
      ___ / /_ ___ ________ __ _
     (_-</ __// -_) __/ __/ _ `/
    /___)⎜__/ \__/_/ /_/  \_,_/
    https://github.com/novitae
    ⎺⎺⎺⎺⎺⎺⎺⎺⎺⎺⎺⎺⎺ ⎺⎺⎺⎺⎺⎺ ⎺⎺ ⎺"""
    _LOGO = _LOGO[:VERSION_POSITON]+cs(VERSION,OTHER).underline().bold()+_LOGO[VERSION_POSITON+len(VERSION):]

    for char in "(/|,)⎜\\<`_-⎺":
        _LOGO = _LOGO.replace(char,str(cs(char, HORIZONTAL if char in "_-⎺" else VERTICAL).bold()))
    return _LOGO.replace(NOVI_URL.replace("/", str(cs("/",VERTICAL).bold())), str(cs(NOVI_URL,OTHER).underline()))

def _return_infos() -> str:
    def random_col(s) -> str:
        return "".join([str(cs(l,choice([HORIZONTAL,"SteelBlue3","SteelBlue4","SlateBlue4"]))) for l in s])
    L = []
    messages = str(f"""
     {cs("A SOCMINT tool for Instagram using follow lists;",OTHER).underline()}
    * {random_col("Export followers / following / mutuals of your target")}
    * {random_col("Compare lists to find mutuals, non-mutuals and evolutions")}
    * {random_col("Lists analysis to get close circle accounts probabilites")}""").split("\n")
    for i, line in enumerate(_generate_logo().split("\n")):
        try:
            L.append(line+messages[i].replace("*",str(cs("*",OTHER))))
        except IndexError:
            L.append(line)

    return "\n".join(L)

LOGO = _generate_logo()


from json import dumps
from os import path
from stringcolor import *

# INFOS -------------------------------------------------------
VERSION = "2.2"
PROG = "sterra"
USAGE = ""
DESCRIPTION = "Sterra is a SOCMINT tool to analyse instagram followers|following of a target, and get informations from it."
EPILOG = "WIKI: "+str(underline("https://github.com/novitae/SterraDev/wiki"))

DAY_STRF = "%b-%d-%Y"
HOUR_STRF = "%H-%M-%S"

# PARSER ------------------------------------------------------
HISTORY_STORE_TRUE = ["all","clear","clear-parts"]
HISTORY_ARGUMENTS = {
    "all":"show all the export history",
    "clear":"clear the history",
    "clear-parts":"delete all part stored",
    "compare-tree":"shows the tree of a compare file",
    "file-id":"show the path associated to the filled id",
    "match":"show the items containing the filled string",
    "path":"show the id associated to the filled path"
    }
HISTORY_ARG_KEYS = [x.replace("-","_") for x in list(HISTORY_ARGUMENTS.keys())]
CLEARED_HISTORY = dumps({"part":{}},indent=4)

# USERNAME LISTS ----------------------------------------------
def makeMutuals(t:tuple) -> list:
    """Returns username in both lists contained in the input tuple"""
    l1, l2 = t
    _max = l1 if len(l1) >= len(l2) else l2
    _min = l1 if _max == l2 else l2
    return [u for u in _min if u in _max]

# PATH / NAMES ------------------------------------------------
MODULE_PATH = path.dirname(__file__)

DEFAULT_EXPORT_PATH = path.join(MODULE_PATH,"export/")
PARTS_PATH = path.join(MODULE_PATH,"parts/")
HISTORY_PATH = path.join(MODULE_PATH,'_history_.json')

PATH_TO_CHECK = [DEFAULT_EXPORT_PATH,PARTS_PATH]

# OTHER --------------------------------------------------------
ENDPOINTS_TEST_LIST = [
    'accounts/edit/?__a=1',
    'accounts/manage_access/?__a=1',
    'push/web/settings/?__a=1',
    'session/login_activity/?__a=1',
    'emails/emails_sent/?__a=1',
    'settings/help/?__a=1',
    'accounts/convert_to_professional_account/?__a=1'
    ]


from sterra.sterools import makeMutuals, isId, checkPaths
from sterra._logo_ import LOGO
from sterra._outputs_ import _ as outputs
from sterra._sterrage_ import DEFAULT_EXPORT_PATH, HISTORY_ARG_KEYS, PROG, DESCRIPTION, EPILOG, HISTORY_ARGUMENTS, HISTORY_STORE_TRUE, HISTORY_PATH, PARTS_PATH
from sterra.SterraException import RateLimitError, BadListOfDictError, ExportDataError

from argparse import ArgumentParser, SUPPRESS
from stringcolor import *

from sterra.filerra import *
from sterra.stegram import _instagram
from sterra.comparerra import _compare
from sterra.stanalysis import _analysis
from sterra.exterra import exman
import sterra.histerra as hst
import sys
import json

#+ -> to do
#! -> vulnerable point

global _
_ = outputs()

def EXPORT(**kwargs) -> bool:
    """Exports the follow list targeted and the information of the accounts in it"""
    def autoLaunchBack(flist:list) -> list: # Not used for now
        """Relaunch the process everytime it gets blocked until the work is done"""
        lback_scraped = already if already else []
        while len(lback_scraped) != len(flist):
            lback_scraped = scraper.getUsernameDetails(flist, already=lback_scraped)
            scraper.distractAPI()
        return lback_scraped

    def scrape(targlist:str,flist:list) -> None:
        """Scrapes the data of usernames in flist and returns the id of the file in wich it has been written"""
        if kwargs.get("only_usernames"):
            _.p(f"""{username}'s {targlist} usernames file exported under id: {str(bold(exporter([{"username":usn} for usn in flist],_,**kwargs)()))}""",logo=logo_to_choose)

        else:
            # if kwargs.get("auto_launckback"):
            #     __a1List = autoLaunchBack(flist)
            # else:
                __a1List = scraper.getUsernameDetails(flist, already=already)
                
                if len(__a1List) != len(flist): # Export en part
                    value = exman(_).part(targlist, part_path)
                    try:
                        pid, file_path, _t = value
                    except ValueError:
                        pid, file_path = value
                    exporter(_, __a1List, file_path, Format="json")()
                    _.r(RateLimitError(f"Part file exported under id: {str(bold(pid))}"),cont=True)
                    _.p("Check this link for more -> https://github.com/novitae/sterraxcyl/wiki/Export#-parts",logo="Interrogation",ext=True)
                
                else: # Export classique
                    fid, file_path = exman(_).custom_name(name=kwargs["name"], Format=kwargs["format"], path=kwargs["path"]) if kwargs.get("name") else exman(_).classic_export(target=targlist, username=username, Format=kwargs["format"], path=kwargs["path"])
                    exporter(_=_, List=__a1List, file_path=file_path, Format=Format, **{"username":username,"target":targlist})()
                    _.p(f"""{username}'s {targlist} list exported under id: {str(bold(fid))}""",logo=logo_to_choose)
    
    # Options de parts
    part_option = kwargs.get("part")
    part_path = None

    if part_option:
        _i, part_path, target_list = hst.get_part(part_option)
        kwargs["target"] = target_list
        _.p("Reading part content ...",logo="json")
        part_content = reader(_, part_path)()
    else:
        target_list = kwargs["target"]

    already = part_content if part_option else None
    original_target = kwargs["target"]
    logo_to_choose = Format = kwargs["format"]
    scraper = _instagram(_, original_target, **kwargs)
    
    lists = scraper.scrapeTargetLists()
    kwargs["username"] = username = scraper.username

    follow_list = (
        {target_list:lists[0]} if target_list in ["followers","following"]
        else {"followers":lists[0],"following":lists[1]} if target_list == "both"
        else {"mutuals":makeMutuals(lists)}
        )

    for t, l in follow_list.items():
        scrape(t,l)

    return True

def COMPARE(**kwargs):
    """Compare lists between them to return commons or differences"""
    Format = logo_to_choose = kwargs.get("format")
    no_print = kwargs.get("no_print")
    fi = kwargs.get("fi")

    ordre = (1, 0) if kwargs.get("invert") else (0, 1)
    f1p, f2p = fi[ordre[0]], fi[ordre[1]]
    l1, l2 = reader(_=_, file_path=f1p)(), reader(_=_, file_path=f2p)()

    c = _compare(_, l1, l2)
    if not no_print:
        _.p(f"""Between {underline(f1p.split("/")[-1])} and {underline(f2p.split("/")[-1])},""",logo="Compare")
    for targ_comp in ["common_usernames","not_common_usernames"]:
        if kwargs.get(targ_comp):
            r = getattr(c,targ_comp)()
            if not no_print:
                _.p(f"""    {bold(targ_comp.capitalize().replace("_"," "))}:""")
                if r:
                    for item in r:
                        _.p("    -> "+(underline("".join([str(cs(item, color)) for item, color in {"https://www.instagram.com/":'LightGrey13',item["username"]:'LightGrey5',"/":'LightGrey13'}.items()])) if kwargs.get("url") else item["username"]))
                else:
                    _.p(f"""    -> No {targ_comp.replace("_"," ")} found.""")

            if kwargs.get("export"):
                if r:
                    fid, file_path = exman(_).custom_name(name=kwargs["name"], Format=Format, path=kwargs["path"]) if kwargs.get("name") else exman(_).compare_export(file1=f1p, file2=f2p, target=targ_comp, Format=Format, path=kwargs["path"])
                    exporter(_=_, List=r, file_path=file_path, Format=Format, **{"target":targ_comp})()
                    _.p(f"""Comparaison exported under id: {str(bold(fid))}""",logo=logo_to_choose)
                else:
                    _.r(ExportDataError("the compared result is empty"))
    return True

def ANALYSE(**kwargs) -> bool:
    def arrow_color(i:float) -> str:
        colors = ['Black','DarkRed','DarkRed2','Red4','Red3','Red2','Red','OrangeRed','DarkOrange','Orange','LightGoldenrod2','Yellow','Yellow2','Yellow3','Yellow4','GreenYellow','Chartreuse2','Green2','Green3','Green4']
        return str(cs("->",colors[int(i/5)]))

    Format = logo_to_choose = kwargs.get("format")
    path = kwargs["fi"]
    data = reader(_=_, file_path=path)()
    
    required_keys = ['biography',"full_name",'is_private','followers','following','external_url','is_verified','posts','is_professional_account','is_business_account',"id"]
    missing_keys = [k for k in required_keys if k not in list(data[0].keys())]
    if missing_keys:
        _.r(BadListOfDictError(missing_keys[0]))    

    for analyse_targ in ["personnal","interests"]:
        if kwargs.get(analyse_targ):
            analysed = _analysis(_=_, data=data, **kwargs)(analyse_targ)
            # example # analysed = [{"username":"username","%":x*5} for x in range(20)]
            if not kwargs.get("no_print"):
                _.p(f"""Probabilites for {bold(analyse_targ.replace("_"," "))} accounts in {underline(path.split("/")[-1])}:\n    %% -> username""", logo="Analyse")
                for account in analysed:
                    _.p(f"""    {str(
                        str(int(account["%"])))+" " if account["%"] < 10 else str(int(account["%"]))} {arrow_color(account["%"])} {"".join([str(cs(item, color)) for item, color in {"https://www.instagram.com/":'LightGrey13',account["username"]:'LightGrey5',"/":'LightGrey13'}.items()]
                        ) if kwargs.get("url") else account["username"]}""")

            if kwargs.get("export"):
                list_id = hst.path(**{"path":path})
                fid, file_path = exman(_).custom_name(name=kwargs["name"], Format=Format, path=kwargs["path"]) if kwargs.get("name") else exman(_).analysis_export(list_id=list_id, target=analyse_targ, Format=Format, path=kwargs["path"])
                exporter(_=_, List=analysed, file_path=file_path, Format=Format, **kwargs)()
                _.p(f"""Analyse exported under id: {str(bold(fid))}""",logo=logo_to_choose)

    return True

def CONVERT(**kwargs) -> bool:
    infoList = reader(_, kwargs["fi"])()
    infos =  exman(_)._decompose_path(kwargs["fi"])
    name = kwargs["name"] if kwargs.get("name") else infos["name"]
    fid, file_path = exman(_).custom_name(name=name, Format=kwargs["format"], path=kwargs["path"])
    exporter(_=_, List=infoList, file_path=file_path, Format=kwargs["format"])()
    _.p(f"""Converted file exported under id: {str(bold(fid))}""", logo=kwargs["format"])
    return True

def HISTORY(**kwargs:dict) -> None:
    funcKey = list(kwargs.keys())[0]
    kwargs["tostr"] = True
    respnse = getattr(hst,funcKey)(**kwargs)
    if type(respnse) is tuple:
        rep, logo = respnse
        _.p(rep, logo=logo)
    else:
        if type(respnse) is dict:
            toPr = ("\n    ".join([f"""{k} -> {json.dumps(v) if k == "part" else v["path"]}""" for k, v in respnse.items()]))
            if funcKey == "all":
                _.p(f"Full history:\n    {toPr}", logo="i")
        else:
            _.p(respnse, logo="i")
    return None

def _parser() -> tuple[str, dict]:
    ARG_FORMAT = "format of the file (default:\"excel\")"
    ARG_EXPORT_PATH = f"""path to the directory to export the file(s) (default:"{DEFAULT_EXPORT_PATH}")"""
    ARG_SELECTED_FILE = "id or path of the file"
    ARG_URL = "print url instead of usernames"

    def _formatDefault(a:dict) -> dict:
        """Capture les erreurs d'arguments, change les id de path en path."""
        _.colors = False if a.get("no_colors") else True
        _._raise = a.get("raw_raising")

        wich = a["wich"]
        no_false = {k:v for k,v in a.items() if (k if wich == "analyse" else v)} # Analyse module needs all value including None or False
        nfKeys = list(no_false.keys())

        if wich in ["compare","analyse"]:
            if wich == "compare":
                compare_requ_args = ["not_common_usernames","common_usernames"]
                if not set(compare_requ_args) & set(list(no_false.keys())):
                    parser.error(f"""compare: compare requires at least one of the arguments: {", ".join([f'''--{x.replace("_","-")}''' for x in compare_requ_args])}""")

            elif wich == "analyse":
                analyse_requ_args = ["personnal","interests"]
                if not [item for item in list(no_false.keys()) if a[item] and item in analyse_requ_args]:
                    no_false["personnal"] = True

            if wich in ["compare","analyse"]:
                if a["no_print"] and not a["export"]:
                    parser.error(f"{wich}: {wich} requires --export if you choose --no-print (avoids making a job without any output)")

        if wich == "history":
            if len([k for k, v in no_false.items() if k in HISTORY_ARG_KEYS]) != 1:
                parser.error("history: history module requires only one argument between the ones available")
        
        if no_false.get("no_exp_limit") and not a.get("express"):
            parser.error("export: --no-exp-limit can't be applied if --express is not enabled")
        
        if "fi" in nfKeys:
            if type(no_false["fi"]) is list:
                no_false["fi"] = [(hst.file_id(fi) if isId(fi) else fi) for fi in no_false["fi"]]
                if None in no_false["fi"]:
                    _.r(IndexError('One of the input fileId have not been found in the history.'))
            else:
                no_false["fi"] = hst.file_id(no_false["fi"]) if isId(no_false["fi"]) else no_false["fi"]
                if not no_false["fi"]:
                    _.r(IndexError('The input fileId have not been found in the history.'))
        
        if "format" in nfKeys:
            no_false["format"] = no_false["format"] if no_false["format"] != "excel" else "xlsx"

        if "name" in nfKeys:
            classic_err = "--name can't be used to export two files (they would have the same name)"
            if wich == "export":
                if no_false.get("target") == "both":
                    parser.error(classic_err)
            
            elif wich == "compare":
                if "not_common_usernames" in nfKeys and "common_usernames" in nfKeys and "export" in nfKeys:
                    parser.error(classic_err)

            elif wich == "analyse":
                if a["personnal"] and a["interests"] and a["export"]:
                    parser.error(classic_err)

        no_false.pop("wich")
        return no_false

    parser = ArgumentParser(prog=PROG,description=DESCRIPTION,epilog=EPILOG)
    
    parser.add_argument("--raw-raising", action="store_true",
                        help="raises Exceptions with python \"raise\"; useful for debug.")
    parser.add_argument("--no-colors", action="store_true",
                        help="print all outputs in b&w")

    sub_parsers = parser.add_subparsers()

    # export
    export_parser = sub_parsers.add_parser("export", add_help=False)
    
    exp_required = export_parser.add_argument_group(title="positional arguments")
    exp_exclued_username = exp_required.add_mutually_exclusive_group(required=True)
    exp_exclued_username.add_argument("-u", "--username", metavar="U", type=str,
                        help="target's instagram username")
    exp_exclued_username.add_argument("-id", metavar="ID", type=int,
                        help="target's instagram id")
    exp_exclued_target = exp_required.add_mutually_exclusive_group(required=True)
    exp_exclued_target.add_argument("-t", "--target", choices=["followers", "following", "both", "mutuals"],
                        help="list to export")
    exp_exclued_target.add_argument("-p", "--part", const=True, nargs="?",
                        help="id of the part") # The part id leads to a dict in history containing its path and the target list
    
    exp_login = export_parser.add_argument_group(title="login arguments")
    exp_exc_login = exp_login.add_mutually_exclusive_group(required=True)
    exp_exc_login.add_argument("-lcrd", "--login-credentials", metavar=("U", "P"), nargs=2,
                        help="login by credentials (USERNAME PASSWORD)")
    exp_exc_login.add_argument("-ssid", "--login-session-id", metavar="S",
                        help="login by sessionid")

    exp_export = export_parser.add_argument_group(title="export arguments")
    exp_export_data = exp_export.add_mutually_exclusive_group()
    exp_export_data.add_argument("--all-infos", action="store_true",
                        help="exports extra informations (not interesting ones)")
    exp_export_data.add_argument("--only-usernames", action="store_true",
                        help="exports only the usernames, not the data linked to")
    exp_export.add_argument("-f", "--format", choices=["excel", "csv", "json"], default="excel",
                        help=ARG_FORMAT)
    exp_export.add_argument("--path", metavar="P", default=DEFAULT_EXPORT_PATH,
                        help=ARG_EXPORT_PATH)
    exp_export.add_argument("--name", metavar="N",
                        help="custom name for the exported file")

    exp_speed = export_parser.add_argument_group(title="speed arguments")
    exp_exc_speed = exp_speed.add_mutually_exclusive_group()
    exp_exc_speed.add_argument("-d", "--delay", type=int,
                        help="delay between requests")
    exp_exc_speed.add_argument("-e", "--express", action="store_true",
                        help="express requests (deactivated if > 109 total usernames to avoid blocking)")

    exp_optional = export_parser.add_argument_group(title="optional arguments")
    exp_optional.add_argument("--no-exp-limit", action="store_true",
                        help="disable the username count limitation on express mode")
    # For the day instagram will find a way to put a more frequent ratelimit
    #exp_optional.add_argument("--auto-launchback", action="store_true",
    #                    help="relaunch automatically the scraping while all usernames haven't been scraped")
    exp_optional.add_argument('-h', '--help', action='help', default=SUPPRESS,
                        help='show this help message and exit')
    
    export_parser.set_defaults(wich="export")

    # compare
    compare_parser = sub_parsers.add_parser("compare")

    compare_parser.add_argument("fi", metavar="{ID,PATH} {ID,PATH}", nargs=2,   # Must do a metavar like this to avoid:
                        help=f"{ARG_SELECTED_FILE}s to compare")                # https://stackoverflow.com/questions/20680882/in-python-argparse-crashes-when-using-h
    compare_parser.add_argument("-n", "--no-print", action="store_true",
                        help="doesn\"t print results")
    compare_parser.add_argument("-u", "--url", action="store_true",
                        help=ARG_URL)
    compare_parser.add_argument("-i", "--invert", action="store_true",
                        help="invert the order of the lists")

    com_want = compare_parser.add_argument_group(title="action")
    com_want.add_argument("--common-usernames", action="store_true",
                        help="commons usernames between lists")
    com_want.add_argument("--not-common-usernames", action="store_true",
                        help="not commons usernames between lists (result vary depending on the list order; use --invert)")

    comp_export = compare_parser.add_argument_group(title="export data")
    comp_export.add_argument("-e", "--export", action="store_true",
                        help="export the result")
    comp_export.add_argument("-f", "--format", choices=["excel", "csv", "json"], default="excel",
                        help=ARG_FORMAT)
    comp_export.add_argument("-p", "--path", default=DEFAULT_EXPORT_PATH,
                        help=ARG_EXPORT_PATH)
    comp_export.add_argument("--name", metavar="N",
                        help="custom name for the exported file")

    compare_parser.set_defaults(wich="compare")

    # analyse
    analyse_parser = sub_parsers.add_parser("analyse")

    analyse_parser.add_argument("fi", metavar="{ID,PATH}",
                        help=f"{ARG_SELECTED_FILE} to analyse")
    analyse_parser.add_argument("-i", "--invert", action="store_true",
                        help="print by the end of the list")
    analyse_parser.add_argument("-e", "--export", action="store_true",
                        help="export results")
    analyse_parser.add_argument("-f", "--format", choices=["excel", "csv", "json"], default="excel",
                        help=ARG_FORMAT)
    analyse_parser.add_argument("--ignore-over", type=int,
                        help="quantity of followers over wich we consider the account as impossible to be in the close circle of the target (default:10000)")
    analyse_parser.add_argument("--name", metavar="N",
                        help="custom name for the exported file")
    analyse_parser.add_argument("--no-print", action="store_true",
                        help="don't print results")
    analyse_parser.add_argument("-p", "--path", default=DEFAULT_EXPORT_PATH,
                        help=ARG_EXPORT_PATH)
    analyse_parser.add_argument("--pctg", type=int,
                        help="percentage under wich results will not be printed (between 0 and 98)")
    analyse_parser.add_argument("-s", "--size", type=int,
                        help="size of the output")
    analyse_parser.add_argument("-u", "--url", action="store_true",
                        help=ARG_URL)

    ana_want = analyse_parser.add_argument_group(title="analysis")
    ana_want.add_argument("--personnal", action="store_true",
                        help="probabilities for close circle accounts")
    ana_want.add_argument("--interests", action="store_true",
                        help="probabilities for interests account")

    analyse_parser.set_defaults(wich="analyse")

    # convert
    convert_parser = sub_parsers.add_parser("convert")
    
    convert_parser.add_argument('fi', metavar="{ID,PATH}",
                        help=f'{ARG_SELECTED_FILE} to convert')
    convert_parser.add_argument('-f', '--format', choices=['excel', 'csv', 'json'], required=True,
                        help=ARG_FORMAT)
    convert_parser.add_argument("--name", metavar="N",
                        help="custom name for the exported file")
    convert_parser.add_argument("-p", "--path", metavar="P", default=DEFAULT_EXPORT_PATH,
                        help=ARG_EXPORT_PATH)

    convert_parser.set_defaults(wich="convert")

    # history
    history_parser = sub_parsers.add_parser("history", usage=f"""sterra history [-h] {{{"|".join([(f"--{a}" if a in HISTORY_STORE_TRUE else f"--{a} {a.upper()}") for a in HISTORY_ARGUMENTS])}}}""")
    
    history_parser.add_argument("-a", f"--all", action="store_true",
                        help=f"""show all the export history""")
    history_parser.add_argument(f"--clear", action="store_true",
                        help=f"""clear the history""")
    history_parser.add_argument(f"--clear-parts", action="store_true",
                        help=f"""delete all part stored""")
    history_parser.add_argument(f"--compare-tree",
                        help=f"""shows the tree of a compare file""")
    history_parser.add_argument("-i", f"--file-id",
                        help=f"""show the path associated to the filled id""")
    history_parser.add_argument("-m", f"--match",
                        help=f"""show the items containing the filled string (can be a regex if it is used like ' ?"YOUR REGEX"? ')""")
    history_parser.add_argument("-p", f"--path",
                        help=f"""show the id associated to the filled path""")
    
    history_parser.set_defaults(wich="history")

    args = parser.parse_args()
    
    try:
        return args.wich, _formatDefault(vars(args))
    except AttributeError:
        parser.exit(parser.print_help())

def main():
    if len(sys.argv) > 1 and sys.argv[1] == "uninstall":
        _.p("To uninstall properly sterra, run the following commands:",logo="Python")
        _.p(f" $ sudo rm -rf {DEFAULT_EXPORT_PATH}")
        _.p(f" $ sudo rm -rf {PARTS_PATH}")
        _.p(f" $ sudo rm -rf {HISTORY_PATH}")
        _.p(f" $ pip uninstall sterra")
    
    else:
        module, kwargs = _parser()
        _.p(LOGO)
        ret = globals()[module.upper()](**kwargs)
        if ret:
            _.p("If you have ideas to improve this program, don't hesitate to write them in the issue section !",logo="Plus")

if __name__ == "__main__":
    main()


class parentException(Exception):
    def __init__(self, *args: object) -> None:
        self.args = args
        super().__init__(*args)

    def __str__(self) -> str:
        return ": ".join(self.args)

#+ essayer de formuler les erreurs avec de la couleur avant qu'elles ne soient print

# Global
class RateLimitError(parentException):
    def __init__(self,msg:str="") -> None:
        super().__init__(f"""Rate limit {f"at {msg}" if msg else msg} blocks the program. Wait a bit, or change of login account.""")

# Defaut
class UserNotFoundError(parentException):
    def __init__(self,userPointer:str) -> None:
        super().__init__(f"User not found. Verify its {userPointer}.")
class PrivateAccError(parentException):
    def __init__(self,u:str=None) -> None:
        super().__init__((u if u else "Your target")+" is a private account that you're not following.")
class NoFollowError(parentException):
    def __init__(self,m:str=None) -> None:
        super().__init__(m if m else "Your target has not followers nor following.")
class MutualsError(parentException):
    def __init__(self,m:str) -> None:
        super().__init__(m)
class BadListOfDictError(parentException):
    def __init__(self,m:str) -> None:
        super().__init__(f"The input list doesn't contain the key \"{m}\" used for the analysis.")
class NoSuchLogoError(parentException):
    def __init__(self,m:str) -> None:
        super().__init__(f"The logo \"{m}\" doesn't exist.")
class ExportDataError(parentException):
    def __init__(self,m:str) -> None:
        super().__init__(f"The file can't be exported; {m}.")

# Custom
class LoginError(parentException):
    pass


from os import path, makedirs
from sterra._sterrage_ import PATH_TO_CHECK, HISTORY_PATH, CLEARED_HISTORY

# USERNAME LISTS ----------------------------------------------
def makeMutuals(t:tuple) -> list: #
    """Returns username in both lists contained in the input tuple"""
    l1, l2 = t
    _max = l1 if len(l1) >= len(l2) else l2
    _min = l1 if _max == l2 else l2
    return [u for u in _min if u in _max]

# PATH / NAMES ------------------------------------------------
def checkPaths() -> list: #
    rtr = []
    for p in PATH_TO_CHECK:
        if not path.exists(p):
            makedirs(p)
            rtr.append(f"Path {p} created.")
    
    try:
        open(HISTORY_PATH,"r").close()
    except FileNotFoundError:
        rtr.append(f"Path {HISTORY_PATH} created.")
        open(HISTORY_PATH,"w").write(CLEARED_HISTORY)

    return rtr

def isId(p:str) -> bool: #
    """Returns True if id, False if path"""
    try:
        int(p)
        return True
    except ValueError:
        return False